from pydantic_client.decorators import delete, get, patch, post, put
from pydantic_client.main import PydanticClientManager
from pydantic_client.schema.client_config import ClientConfig, ClientType

pydantic_client_manager = PydanticClientManager()

__all__ = [
    "pydantic_client_manager",
    "ClientConfig",
    "ClientType",
    "patch",
    "get",
    "post",
    "put",
    "delete",
]



import logging

from pydantic_client.container import container
from pydantic_client.schema.client_config import ClientConfig

logger = logging.getLogger(__name__)


class PydanticClientManager:
    def __init__(self):
        # proto_class: protocol_object
        self.pydantic_clients = {}

    def register(
        self,
        client_config: ClientConfig
    ):
        def wrapper(protocol_class):
            web_client = client_config.get_client()
            protocol = protocol_class()
            container.bind_protocol(protocol, web_client)
            if protocol_class in self.pydantic_clients:
                logger.warning(
                    f"protocol {protocol_class} already registered, will be overwritten")
            self.pydantic_clients[protocol_class] = protocol
            return protocol_class

        return wrapper

    def get(self, protocol_class=None):
        if len(self.pydantic_clients) == 1:
            return self.pydantic_clients[list(self.pydantic_clients.keys())[0]]
        if not protocol_class:
            raise ValueError(
                "Multiple clients exists, protocol_class is required")
        client = self.pydantic_clients.get(protocol_class)
        if not client:
            raise ValueError(f"client for {protocol_class} not found")
        return client



from typing import Any

from pydantic_client.clients.abstract_client import AbstractClient
from pydantic_client.schema.http_request import HttpRequest

try:
    from httpx import AsyncClient
except ImportError:
    raise ImportError("Please install `httpx` to use HttpxClient")


class HttpxClient(AbstractClient):

    def get_session(self):
        session = super().get_session()
        return session if isinstance(session, AsyncClient) \
            else AsyncClient(http2=self.config.http2)

    async def do_request(self, request: HttpRequest) -> Any:
        async with self.get_session() as session:
            try:
                response = await session.request(**self.parse(request))
                response.raise_for_status()
                if response.is_success:
                    if not request.is_file:
                        return response.json()
                    return response.read()
            except BaseException as e:
                raise e



from typing import Any, Dict

from pydantic_client.schema.client_config import ClientConfig
from pydantic_client.schema.http_request import HttpRequest


class AbstractClient:

    def __init__(self, config: ClientConfig):
        self.config = config
        self.base_url = config.base_url.rstrip("/")

    def get_session(self):
        return self.config.client_session

    def do_request(self, request: HttpRequest) -> Any:
        raise NotImplementedError

    def parse(self, request: HttpRequest) -> Dict[str, Any]:
        if request.data:
            data, json = request.data, None
        else:
            data, json = None, request.json_body
        headers = request.request_headers if request.request_headers \
            else self.config.headers
        url = self.base_url + request.url
        return dict(url=url, data=data, json=json, headers=headers,
                    method=request.method, timeout=self.config.timeout)

    @classmethod
    def from_toml(cls, toml_config: ClientConfig):
        return cls(toml_config)



from typing import Any

from pydantic_client.clients.abstract_client import AbstractClient
from pydantic_client.schema.http_request import HttpRequest

try:
    from requests import Session
except ImportError:
    raise ImportError("Please install `requests` to use RequestsClient")


class RequestsClient(AbstractClient):
    session = Session()

    def get_session(self) -> Session:
        session = super().get_session()
        return session if isinstance(session, Session) else self.session

    def do_request(self, request: HttpRequest) -> Any:
        try:
            response = self.get_session().request(**self.parse(request))
            response.raise_for_status()
            return response.json() if not request.is_file else response.content
        except BaseException as e:
            raise e



from typing import Any, Callable

from pydantic_client.clients.abstract_client import AbstractClient
from pydantic_client.schema.http_request import HttpRequest

try:
    from aiohttp.client import ClientSession
except ImportError:
    raise ImportError("Please install `aiohttp` to use AIOHttpClient")


class AIOHttpClient(AbstractClient):

    def get_session(self) -> Callable[[], ClientSession]:
        session = super().get_session()
        return lambda: ClientSession() if not session else session()

    async def do_request(self, request: HttpRequest) -> Any:
        session_factory = self.get_session()
        s = session_factory()
        async with s as session:
            try:
                req = session.request(**self.parse(request))
                async with req as resp:
                    resp.raise_for_status()
                    if resp.status == 200:
                        if not request.is_file:
                            return await resp.json()
                        return await resp.content.read()
            except BaseException as e:
                raise e



from typing import Any, Callable, Dict, Optional

from pydantic import BaseModel

from pydantic_client.schema.file import File


class MethodInfo(BaseModel):
    func: Callable
    http_method: str
    url: str
    request_type: Dict[str, Any]
    response_type: Optional[Any]
    form_body: bool
